#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>

#include "libkvm.h"
#include "runtime.h"

/* callback definitions as shown in Listing 2 go here */

void load_file(void *mem, const char *filename)
{
    int  fd;
    int  nr;

    fd = open(filename, O_RDONLY);
    if (fd == -1) {
        fprintf(stderr, "Cannot open %s", filename);
        perror("open");
        exit(1);
    }
    while ((nr = read(fd, mem, 4096)) != -1  &&  nr != 0)
        mem += nr;

    if (nr == -1) {
        perror("read");
        exit(1);
    }
    close(fd);
}

#define MEMORY_SIZE     (0x1000000)     /* 16 Mb */
#define FIRST_VCPU      (0)

static int my_inb(void *opaque, uint16_t addr, uint8_t *data)
{ puts ("inb"); return 0; }

static int my_inw(void *opaque, uint16_t addr, uint16_t *data)
{ puts ("inw"); return 0; }

static int my_inl(void *opaque, uint16_t addr, uint32_t *data)
{ puts ("inl"); return 0; }

static int my_outb(void *opaque, uint16_t addr, uint8_t data)
{ puts ("outb"); return 0; }

static int my_outw(void *opaque, uint16_t addr, uint16_t data)
{ puts ("outw"); return 0; }

static int my_outl (void *opaque, uint16_t addr, uint32_t data)
{ puts ("outl"); return 0; }

static int my_mmio_read (void *opaque, uint64_t addr, uint8_t *data, int len) 
{ puts ("my_mmio_read");  return 0; }

static int my_mmio_write (void *opaque, uint64_t addr, uint8_t *data, int len) 
{ puts ("mmio_write");  return 0; }

static int my_debug (void *opaque, void *env, struct kvm_debug_exit_arch *arch_info) 
{ puts ("debug");  return 0; }

static int my_halt (void *opaque, int vcpu) 
{ puts ("halt");  return 0; }

static int my_shutdown (void *opaque, void *env)
{ puts ("shutdown");  return 0; }

static int my_io_window (void *opaque)
{ puts ("io_window");  return 0; }

static int my_try_push_interrupts (void *opaque)
{ puts ("try_push_interrupts");  return 0; }

static void my_push_nmi (void *opaque) 
{ /* puts ("push_nmi"); */ }

static int my_pre_kvm_run(void *opaque, void * env)
{ /* puts ("pre_kvm_run"); */  return 0; }

static void my_post_kvm_run (void *opaque, void *env) 
{ /* puts ("post_kvm_run"); */  }

static int my_tpr_access (void *opaque, int vcpu, uint64_t rip, int is_write)
{ puts ("tpr_access");  return 0; }

  /*
  ... and similar for my_mmio_read, my_mmio_write,
      my_debug, my_halt, my_shutdown, my_io_window,
      my_try_push_interrupts, my_try_push_nmi,
      my_post_kvm_run, and my_tpr_access
  */

static int my_outb_alt (void *opaque, uint16_t addr, uint8_t data)
{
    #if 1	
    if (addr == IO_PORT_PSEUDO_SERIAL)
        if (isprint(data) || data == '\n')
            putchar(data);
        else
            putchar('.');
    else
    #endif
        printf("outb: %x, %d\n", addr, data);
    fflush (NULL);

    return 0;
}

static struct kvm_callbacks my_callbacks = {
    .inb                 = my_inb,
    .inw                 = my_inw,
    .inl                 = my_inl,
    .outb                = my_outb_alt,
    .outw                = my_outw,
    .outl                = my_outl,
    .mmio_read           = my_mmio_read,
    .mmio_write          = my_mmio_write,
    .debug               = my_debug,
    .halt                = my_halt,
    .shutdown            = my_shutdown,
    .io_window           = my_io_window,
    .try_push_interrupts = my_try_push_interrupts,
    .push_nmi            = my_push_nmi, 
    .post_kvm_run        = my_post_kvm_run,
    .pre_kvm_run         = my_pre_kvm_run,
    .tpr_access          = my_tpr_access
};

int main(int argc, char *argv[])
{
    kvm_context_t  kvm;
    void           *memory_area;

    /* Second argument is an opaque, we don't use it yet */
    kvm = kvm_init(&my_callbacks, NULL);
    if (!kvm) {
        fprintf(stderr, "KVM init failed");
        exit(1);
    }
    if (kvm_create(kvm, MEMORY_SIZE, &memory_area) != 0) {
        fprintf(stderr, "VM creation failed");
        exit(1);
    }
#ifndef KVM_VERSION_LESS_THAN_65
    if (kvm_create_vcpu(kvm, FIRST_VCPU)) {
        fprintf(stderr, "VCPU creation failed");
        exit(1);
    }
#endif
    memory_area = kvm_create_phys_mem(kvm, 0, MEMORY_SIZE, 0, 1);
    load_file(memory_area + 0xf0000, argv[1]);

    kvm_run(kvm, FIRST_VCPU, /* envp = */ NULL);

    return 0;
}
